@@ -3,8 +3,7 @@ Mac OS X .app build command for distutils
 
 Originally (loosely) based on code from py2exe's build_exe.py by Thomas Heller.
 """
-from pkg_resources import require
-require("altgraph", "modulegraph", "macholib")
+from __future__ import print_function
 
 import imp
 import sys
@@ -13,14 +12,24 @@ import zipfile
 import plistlib
 import shlex
 import shutil
+import textwrap
+import pkg_resources
+
+from py2app.apptemplate.setup import main as script_executable
+from py2app.util import mergecopy, make_exec
+
 
 try:
     import sysconfig
 except ImportError:
     sysconfig = None
-from cStringIO import StringIO
 
-from itertools import chain, imap
+try:
+    from cStringIO import StringIO
+except ImportError:
+    from io import StringIO
+
+from itertools import chain
 
 
 from setuptools import Command
@@ -28,9 +37,8 @@ from distutils.util import convert_path
 from distutils import log
 from distutils.errors import *
 
-from altgraph.compat import *
 
-from modulegraph.find_modules import find_modules, parse_mf_results
+from modulegraph.find_modules import find_modules, parse_mf_results, find_needed_modules
 from modulegraph.modulegraph import SourceModule, Package, Script
 from modulegraph import zipio
 
@@ -51,10 +59,18 @@ from py2app import recipes
 from distutils.sysconfig import get_config_var
 PYTHONFRAMEWORK=get_config_var('PYTHONFRAMEWORK')
 
+try:
+    basestring
+except NameError:
+    basestring = str
 
-def get_zipfile(dist):
+
+def get_zipfile(dist, semi_standalone=False):
     if sys.version_info[0] == 3:
-        return "python%d%d.zip"%(sys.version_info[:2])
+        if semi_standalone:
+            return "python%d.%d/site-packages.zip"%(sys.version_info[:2])
+        else:
+            return "python%d%d.zip"%(sys.version_info[:2])
     return getattr(dist, "zipfile", None) or "site-packages.zip"
 
 def framework_copy_condition(src):
@@ -68,6 +84,10 @@ class PythonStandalone(macholib.MachOStandalone.MachOStandalone):
 
     def copy_dylib(self, src):
         dest = os.path.join(self.dest, os.path.basename(src))
+        if os.path.islink(src):
+            dest = os.path.join(self.dest, os.path.basename(os.path.realpath(src)))
+        else:
+            dest = os.path.join(self.dest, os.path.basename(src))
         return self.appbuilder.copy_dylib(src, dest)
 
     def copy_framework(self, info):
@@ -236,7 +256,9 @@ class py2app(Command):
         ("debug-skip-macholib", None,
          "skip macholib phase (app will not be standalone!)"),
         ("arch=", None, "set of architectures to use (fat, fat3, universal, intel, i386, ppc, x86_64; default is the set for the current python binary)"),
-
+        ("qt-plugins=", None, "set of Qt plugins to include in the application bundle (default None)"),
+        ("matplotlib-backends=", None, "set of matplotlib backends to include (default: include entire package)"),
+        ("extra-scripts=", None, "set of scripts to include in the application bundle, next to the main application script"),
         ]
 
     boolean_options = [
@@ -297,6 +319,9 @@ class py2app(Command):
         self.prefer_ppc = False
         self.filters = []
         self.eggs = []
+        self.qt_plugins = None
+        self.matplotlib_backends = None
+        self.extra_scripts = None
 
     def finalize_options (self):
         if not self.strip:
@@ -318,7 +343,7 @@ class py2app(Command):
         self.excludes.add('site')
         if getattr(self.distribution, 'install_requires', None):
             self.includes.add('pkg_resources')
-            self.eggs = require(self.distribution.install_requires)
+            self.eggs = pkg_resources.require(self.distribution.install_requires)
         dylib_excludes = fancy_split(self.dylib_excludes)
         self.dylib_excludes = []
         for fn in dylib_excludes:
@@ -357,7 +382,7 @@ class py2app(Command):
         if self.semi_standalone:
             self.filters.append(not_stdlib_filter)
 
-        if self.iconfile is None and u'CFBundleIconFile' not in self.plist:
+        if self.iconfile is None and 'CFBundleIconFile' not in self.plist:
             # Default is the generic applet icon in the framework
             iconfile = os.path.join(sys.prefix, 'Resources', 'Python.app',
                 'Contents', 'Resources', 'PythonApplet.icns')
@@ -367,12 +392,16 @@ class py2app(Command):
 
         self.runtime_preferences = list(self.get_runtime_preferences())
 
+        self.qt_plugins = fancy_split(self.qt_plugins)
+        self.matplotlib_backends = fancy_split(self.matplotlib_backends)
+        self.extra_scripts = fancy_split(self.extra_scripts)
+
 
         if self.datamodels:
-            print "WARNING: the datamodels option is deprecated, add model files to the list of resources"
+            print("WARNING: the datamodels option is deprecated, add model files to the list of resources")
 
         if self.mappingmodels:
-            print "WARNING: the mappingmodels option is deprecated, add model files to the list of resources"
+            print("WARNING: the mappingmodels option is deprecated, add model files to the list of resources")
 
 
     def get_default_plist(self):
@@ -387,7 +416,7 @@ class py2app(Command):
             except ValueError:
                 pass
 
-        if not isinstance(version, (str, unicode)):
+        if not isinstance(version, basestring):
             raise DistutilsOptionError("Version must be a string")
 
         if sys.version_info[0] > 2 and isinstance(version, type('a'.encode('ascii'))):
@@ -416,7 +445,8 @@ class py2app(Command):
             # We're in a virtualenv environment, locate the real prefix
             fn = os.path.join(prefix, "lib", "python%d.%d"%(sys.version_info[:2]), "orig-prefix.txt")
             if os.path.exists(fn):
-                prefix = open(fn, 'rU').read().strip()
+                with open(fn, 'rU') as fp:
+                    prefix = fp.read().strip()
 
         try:
             fmwk = macholib.dyld.framework_find(prefix)
@@ -483,7 +513,7 @@ class py2app(Command):
 
 
     def iter_datamodels(self, resdir):
-        for (path, files) in imap(normalize_data_file, self.datamodels or ()):
+        for (path, files) in (normalize_data_file(fn) for fn in (self.datamodels or ())):
             path = fsencoding(path)
             for fn in files:
                 fn = fsencoding(fn)
@@ -496,12 +526,12 @@ class py2app(Command):
 
     def compile_datamodels(self, resdir):
         for src, dest in self.iter_datamodels(resdir):
-            print "compile datamodel", src, "->", dest
+            print("compile datamodel", src, "->", dest)
             self.mkpath(os.path.dirname(dest))
             momc(src, dest)
 
     def iter_mappingmodels(self, resdir):
-        for (path, files) in imap(normalize_data_file, self.mappingmodels or ()):
+        for (path, files) in (normalize_data_file(fn) for fn in (self.mappingmodels or ())):
             path = fsencoding(path)
             for fn in files:
                 fn = fsencoding(fn)
@@ -520,7 +550,7 @@ class py2app(Command):
     def iter_data_files(self):
         dist = self.distribution
         allres = chain(getattr(dist, 'data_files', ()) or (), self.resources)
-        for (path, files) in imap(normalize_data_file, allres):
+        for (path, files) in (normalize_data_file(fn) for fn in allres):
             path = fsencoding(path)
             for fn in files:
                 fn = fsencoding(fn)
@@ -535,11 +565,14 @@ class py2app(Command):
             scripts.update([
                 k for k in target.prescripts if isinstance(k, basestring)
             ])
+            if hasattr(target, 'extra_scripts'):
+                scripts.update(target.extra_scripts)
 
+        scripts.update(self.extra_scripts)
         return scripts
 
     def get_plist_options(self):
-        return dict(
+        result = dict(
             PyOptions=dict(
                 use_pythonpath=bool(self.use_pythonpath),
                 site_packages=bool(self.site_packages),
@@ -547,10 +580,12 @@ class py2app(Command):
                 argv_emulation=bool(self.argv_emulation),
                 emulate_shell_environment=bool(self.emulate_shell_environment),
                 no_chdir=bool(self.no_chdir),
-                optimize=self.optimize,
                 prefer_ppc=self.prefer_ppc,
             ),
         )
+        if self.optimize:
+            result['PyOptions']['optimize'] = self.optimize
+        return result
 
     
     def initialize_plist(self):
@@ -568,9 +603,9 @@ class py2app(Command):
                 raise DistutilsOptionError("icon file must exist: %r"
                     % (self.iconfile,))
             self.resources.append(iconfile)
-            plist[u'CFBundleIconFile'] = os.path.basename(iconfile)
+            plist['CFBundleIconFile'] = os.path.basename(iconfile)
         if self.prefer_ppc:
-            plist[u'LSPrefersPPC'] = True
+            plist['LSPrefersPPC'] = True
 
         self.plist = plist
         return plist
@@ -578,9 +613,25 @@ class py2app(Command):
     def run_alias(self):
         self.app_files = []
         for target in self.targets:
-            dst = self.build_alias_executable(target, target.script)
+
+            extra_scripts = list(self.extra_scripts)
+            if hasattr(target, 'extra_scripts'):
+                extra_scripts.update(extra_scripts)
+
+            dst = self.build_alias_executable(target, target.script, extra_scripts)
             self.app_files.append(dst)
 
+            for fn in extra_scripts:
+                if fn.endswith('.py'):
+                    fn = fn[:-3]
+                elif fn.endswith('.pyw'):
+                    fn = fn[:-4]
+
+                src_fn = script_executable(arch=self.arch)
+                tgt_fn = os.path.join(target.appdir, 'Contents', 'MacOS', fn)
+                mergecopy(src_fn, tgt_fn)
+                make_exec(tgt_fn)
+
     def collect_recipedict(self):
         return dict(iterRecipes())
 
@@ -603,14 +654,18 @@ class py2app(Command):
     def process_recipes(self, mf, filters, flatpackages, loader_files):
         rdict = self.collect_recipedict()
         while True:
-            for name, check in rdict.iteritems():
+            for name, check in rdict.items():
                 rval = check(self, mf)
                 if rval is None:
                     continue
                 # we can pull this off so long as we stop the iter
                 del rdict[name]
-                print '*** using recipe: %s ***' % (name,)
-                self.packages.update(rval.get('packages', ()))
+                print('*** using recipe: %s ***' % (name,))
+
+                if rval.get('packages'):
+                    self.packages.update(rval['packages'])
+                    find_needed_modules(mf, packages=rval['packages'])
+
                 for pkg in rval.get('flatpackages', ()):
                     if isinstance(pkg, basestring):
                         pkg = (os.path.basename(pkg), pkg)
@@ -620,6 +675,12 @@ class py2app(Command):
                 newbootstraps = map(self.get_bootstrap,
                     rval.get('prescripts', ()))
 
+                if rval.get('includes'):
+                    find_needed_modules(mf, includes=rval['includes'])
+
+                if rval.get('resources'):
+                    self.resources.extend(rval['resources'])
+
                 for fn in newbootstraps:
                     if isinstance(fn, basestring):
                         mf.run_script(fn)
@@ -644,12 +705,12 @@ class py2app(Command):
             pdb.post_mortem(sys.exc_info()[2])
 
     def filter_dependencies(self, mf, filters):
-        print "*** filtering dependencies ***"
+        print("*** filtering dependencies ***")
         nodes_seen, nodes_removed, nodes_orphaned = mf.filterStack(filters)
-        print '%d total' % (nodes_seen,)
-        print '%d filtered' % (nodes_removed,)
-        print '%d orphaned' % (nodes_orphaned,)
-        print '%d remaining' % (nodes_seen - nodes_removed,)
+        print('%d total' % (nodes_seen,))
+        print('%d filtered' % (nodes_removed,))
+        print('%d orphaned' % (nodes_orphaned,))
+        print('%d remaining' % (nodes_seen - nodes_removed,))
 
     def get_appname(self):
         return self.plist['CFBundleName']
@@ -660,9 +721,10 @@ class py2app(Command):
             appdir = os.path.join(self.dist_dir, os.path.dirname(base))
             appname = self.get_appname()
             dgraph = os.path.join(appdir, appname + '.html')
-            print ("*** creating dependency html: %s ***"
+            print("*** creating dependency html: %s ***"
                 % (os.path.basename(dgraph),))
-            mf.create_xref(open(dgraph, 'w'))
+            with open(dgraph, 'w') as fp:
+                mf.create_xref(fp)
 
     def build_graph(self, mf, flatpackages):
         for target in self.targets:
@@ -670,9 +732,10 @@ class py2app(Command):
             appdir = os.path.join(self.dist_dir, os.path.dirname(base))
             appname = self.get_appname()
             dgraph = os.path.join(appdir, appname + '.dot')
-            print ("*** creating dependency graph: %s ***"
+            print("*** creating dependency graph: %s ***"
                 % (os.path.basename(dgraph),))
-            mf.graphreport(open(dgraph, 'w'), flatpackages=flatpackages)
+            with open(dgraph, 'w') as fp:
+                mf.graphreport(fp, flatpackages=flatpackages)
 
     def finalize_modulefinder(self, mf):
         for item in mf.flatten():
@@ -682,8 +745,8 @@ class py2app(Command):
                     dn = os.path.dirname(fn)
                     if not os.path.exists(dn):
                         os.makedirs(dn)
-                    fp = open(fn, 'w')
-                    fp.close()
+                    with open(fn, 'w') as fp:
+                        pass
 
                 item.filename = fn
 
@@ -697,10 +760,10 @@ class py2app(Command):
         return py_files, extensions
 
     def collect_packagedirs(self):
-        return filter(os.path.exists, [
+        return list(filter(os.path.exists, [
             os.path.join(os.path.realpath(self.get_bootstrap(pkg)), '')
             for pkg in self.packages
-        ])
+        ]))
 
     def run_normal(self):
         mf = self.get_modulefinder()
@@ -749,7 +812,7 @@ class py2app(Command):
         self.mkpath(self.dist_dir)
 
         self.lib_dir = os.path.join(self.bdist_dir,
-            os.path.dirname(get_zipfile(self.distribution)))
+            os.path.dirname(get_zipfile(self.distribution, self.semi_standalone)))
         self.mkpath(self.lib_dir)
 
         self.ext_dir = os.path.join(self.lib_dir, 'lib-dynload')
@@ -759,7 +822,7 @@ class py2app(Command):
         self.mkpath(self.framework_dir)
 
     def create_binaries(self, py_files, pkgdirs, extensions, loader_files):
-        print "*** create binaries ***"
+        print("*** create binaries ***")
         dist = self.distribution
         pkgexts = []
         copyexts = []
@@ -773,7 +836,7 @@ class py2app(Command):
                     return None
             return fn
         if pkgdirs:
-            py_files = filter(packagefilter, py_files)
+            py_files = list(filter(packagefilter, py_files))
         for ext in extensions:
             fn = packagefilter(ext)
             if fn is None:
@@ -786,7 +849,7 @@ class py2app(Command):
             extmap[fn] = ext
 
         # byte compile the python modules into the target directory
-        print "*** byte compile python files ***"
+        print("*** byte compile python files ***")
         byte_compile(py_files,
                      target_dir=self.collect_dir,
                      optimize=self.optimize,
@@ -803,7 +866,7 @@ class py2app(Command):
 
         # create the shared zipfile containing all Python modules
         archive_name = os.path.join(self.lib_dir,
-                                    os.path.basename(get_zipfile(dist)))
+                                    get_zipfile(dist, self.semi_standalone))
 
         for path, files in loader_files:
             dest = os.path.join(self.collect_dir, path)
@@ -824,8 +887,11 @@ class py2app(Command):
 
         # build the executables
         for target in self.targets:
+            extra_scripts = list(self.extra_scripts)
+            if hasattr(target, 'extra_scripts'):
+                extra_scripts.extend(target.extra_scripts)
             dst = self.build_executable(
-                target, arcname, pkgexts, copyexts, target.script)
+                target, arcname, pkgexts, copyexts, target.script, extra_scripts)
             exp = os.path.join(dst, 'Contents', 'MacOS')
             execdst = os.path.join(exp, 'python')
             if self.semi_standalone:
@@ -834,7 +900,8 @@ class py2app(Command):
                 if os.path.exists(os.path.join(sys.prefix, ".Python")):
                     fn = os.path.join(sys.prefix, "lib", "python%d.%d"%(sys.version_info[:2]), "orig-prefix.txt")
                     if os.path.exists(fn):
-                        prefix = open(fn, 'rU').read().strip()
+                        with open(fn, 'rU') as fp:
+                            prefix = fp.read().strip()
 
                     rest_path = os.path.normpath(sys.executable)[len(os.path.normpath(sys.prefix))+1:]
                     if rest_path.startswith('.'):
@@ -892,7 +959,7 @@ class py2app(Command):
 
         target_dir = os.path.join(target_dir, *(package.identifier.split('.')))
         for dname in package.packagepath:
-            filenames = filter(datafilter, zipio.listdir(dname))
+            filenames = list(filter(datafilter, zipio.listdir(dname)))
             for fname in filenames:
                 if fname in ('.svn', 'CVS'):
                     # Scrub revision manager junk
@@ -900,18 +967,39 @@ class py2app(Command):
                 if fname in ('__pycache__',):
                     # Ignore PEP 3147 bytecode cache
                     continue
+                if fname.startswith('.') and fname.endswith('.swp'):
+                    # Ignore vim(1) temporary files
+                    continue
+                if fname.endswith('~') or fname.endswith('.orig'):
+                    # Ignore backup files for common tools (hg, emacs, ...)
+                    continue
                 pth = os.path.join(dname, fname)
 
                 # Check if we have found a package, exclude those
-                if zipio.isdir(pth) and not pth.endswith('.zip'):
+                if zipio.isdir(pth):
                     # XXX: the 'and not' part is wrong, need to fix zipio.isdir
                     for p in zipio.listdir(pth):
                         if p.startswith('__init__.') and p[8:] in exts:
                             break
 
                     else:
-                        copy_tree(pth, os.path.join(target_dir, fname))
+                        if os.path.isfile(pth):
+                            # Avoid extracting a resource file that happens
+                            # to be zipfile.
+                            # XXX: Need API in zipio for nicer code.
+                            copy_file(pth, os.path.join(target_dir, fname))
+                        else:
+                            copy_tree(pth, os.path.join(target_dir, fname))
                     continue
+
+                elif zipio.isdir(pth) and (
+                        zipio.isfile(os.path.join(pth, '__init__.py'))
+                     or zipio.isfile(os.path.join(pth, '__init__.pyc'))
+                     or zipio.isfile(os.path.join(pth, '__init__.pyo'))):
+                    # Subdirectory is a python package, these will get included later on
+                    # when the subpackage itself is included, ignore for now.
+                    pass
+
                 else:
                     copy_file(pth, os.path.join(target_dir, fname))
 
@@ -928,20 +1016,20 @@ class py2app(Command):
         for dirpath, dnames, fnames in os.walk(self.appdir):
             for nm in list(dnames):
                 if nm.endswith('.dSYM'):
-                    print "removing debug info: %s/%s"%(dirpath, nm)
+                    print("removing debug info: %s/%s"%(dirpath, nm))
                     shutil.rmtree(os.path.join(dirpath, nm))
                     dnames.remove(nm)
         return [file for file in platfiles if '.dSYM' not in file]
 
     def strip_files(self, files):
-        unstripped = 0L
+        unstripped = 0
         stripfiles = []
         for fn in files:
             unstripped += os.stat(fn).st_size
             stripfiles.append(fn)
             log.info('stripping %s', os.path.basename(fn))
         strip_files(stripfiles, dry_run=self.dry_run, verbose=self.verbose)
-        stripped = 0L
+        stripped = 0
         for fn in stripfiles:
             stripped += os.stat(fn).st_size
         log.info('stripping saved %d bytes (%d / %d)',
@@ -1023,7 +1111,19 @@ class py2app(Command):
         indir = os.path.dirname(os.path.join(info['location'], info['name']))
         outdir = os.path.dirname(os.path.join(dst, info['name']))
         self.mkpath(os.path.join(outdir, 'Resources'))
-        pydir = 'python%s'%(info['version'])
+        pydir = 'python%s.%s'%(sys.version_info[:2])
+
+        # Create a symlink "for Python.frameworks/Versions/Current". This
+        # is required for the Mac App-store.
+        os.symlink(
+                os.path.basename(outdir),
+                os.path.join(os.path.dirname(outdir), "Current"))
+
+        # Experiment for issue 57
+        if not os.path.exists(os.path.join(indir, 'include')):
+            alt = os.path.join(indir, 'Versions/Current')
+            if os.path.exists(os.path.join(alt, 'include')):
+                indir = alt
 
         # distutils looks for some files relative to sys.executable, which
         # means they have to be in the framework...
@@ -1113,7 +1213,23 @@ class py2app(Command):
             # bootstrap
             prescripts.append('import_encodings')
 
-        if self.site_packages or self.alias:
+        if os.path.exists(os.path.join(sys.prefix, ".Python")):
+            # We're in a virtualenv, which means sys.path
+            # will be broken in alias builds unless we fix
+            # it.
+            if self.alias or self.semi_standalone:
+                prescripts.append("virtualenv")
+                prescripts.append(StringIO('_fixup_virtualenv(%r)' % (sys.real_prefix,)))
+
+            if self.site_packages or self.alias:
+                import site
+                global_site_packages = not os.path.exists(
+                        os.path.join(os.path.dirname(site.__file__), 'no-global-site-packages.txt'))
+                prescripts.append('virtualenv_site_packages')
+                prescripts.append(StringIO('_site_packages(%r, %r, %d)' % (
+                    sys.prefix, sys.real_prefix, global_site_packages)))
+
+        elif self.site_packages or self.alias:
             prescripts.append('site_packages')
 
         if is_system():
@@ -1127,15 +1243,15 @@ class py2app(Command):
 
         if self.argv_emulation and self.style == 'app':
             prescripts.append('argv_emulation')
-            if u'CFBundleDocumentTypes' not in self.plist:
-                self.plist[u'CFBundleDocumentTypes'] = [
+            if 'CFBundleDocumentTypes' not in self.plist:
+                self.plist['CFBundleDocumentTypes'] = [
                     {
-                        u'CFBundleTypeOSTypes' : [
-                            u'****',
-                            u'fold',
-                            u'disk',
+                        'CFBundleTypeOSTypes' : [
+                            '****',
+                            'fold',
+                            'disk',
                         ],
-                        u'CFBundleTypeRole': u'Viewer'
+                        'CFBundleTypeRole': 'Viewer'
                     },
                 ]
 
@@ -1179,13 +1295,14 @@ class py2app(Command):
         if not isinstance(bootstrap, basestring):
             return bootstrap.getvalue()
         else:
-            return open(bootstrap, 'rU').read()
+            with open(bootstrap, 'rU') as fp:
+                return fp.read()
 
     def create_pluginbundle(self, target, script, use_runtime_preference=True):
         base = target.get_dest_base()
         appdir = os.path.join(self.dist_dir, os.path.dirname(base))
         appname = self.get_appname()
-        print "*** creating plugin bundle: %s ***" % (appname,)
+        print("*** creating plugin bundle: %s ***" % (appname,))
         if self.runtime_preferences and use_runtime_preference:
             self.plist.setdefault(
                 'PyRuntimeLocations', self.runtime_preferences)
@@ -1204,12 +1321,12 @@ class py2app(Command):
         base = target.get_dest_base()
         appdir = os.path.join(self.dist_dir, os.path.dirname(base))
         appname = self.get_appname()
-        print "*** creating application bundle: %s ***" % (appname,)
+        print("*** creating application bundle: %s ***" % (appname,))
         if self.runtime_preferences and use_runtime_preference:
             self.plist.setdefault(
                 'PyRuntimeLocations', self.runtime_preferences)
-        pythonInfo = self.plist.setdefault(u'PythonInfoDict', {})
-        py2appInfo = pythonInfo.setdefault(u'py2app', {}).update(dict(
+        pythonInfo = self.plist.setdefault('PythonInfoDict', {})
+        py2appInfo = pythonInfo.setdefault('py2app', {}).update(dict(
             alias=bool(self.alias),
         ))
         appdir, plist = create_appbundle(
@@ -1240,7 +1357,7 @@ class py2app(Command):
                 basename = fmwk['shortname'] + '.framework'
                 yield os.path.join(fmwk['location'], basename)
     
-    def build_alias_executable(self, target, script):
+    def build_alias_executable(self, target, script, extra_scripts):
         # Build an alias executable for the target
         appdir, resdir, plist = self.create_bundle(target, script)
 
@@ -1257,7 +1374,10 @@ class py2app(Command):
         pyhome = os.path.join(resdir, 'lib', 'python' + sys.version[:3])
         realhome = os.path.join(sys.prefix, 'lib', 'python' + sys.version[:3])
         makedirs(pyhome)
-        self.symlink('../../site.py', os.path.join(pyhome, 'site.py'))
+        if self.optimize:
+            self.symlink('../../site.pyo', os.path.join(pyhome, 'site.pyo'))
+        else:
+            self.symlink('../../site.pyc', os.path.join(pyhome, 'site.pyc'))
         self.symlink(
             os.path.join(realhome, 'config'),
             os.path.join(pyhome, 'config'))
@@ -1294,8 +1414,21 @@ class py2app(Command):
         for fn in target.prescripts:
             bootfile.write(self.get_bootstrap_data(fn))
             bootfile.write('\n\n')
+        bootfile.write("DEFAULT_SCRIPT=%r\n"%(os.path.realpath(script),))
+        script_map = {}
+        for fn in extra_scripts:
+            tgt = os.path.realpath(fn)
+            fn = os.path.basename(fn)
+            if fn.endswith('.py'):
+                script_map[fn[:-3]] = tgt
+            elif fn.endswith('.py'):
+                script_map[fn[:-4]] = tgt
+            else:
+                script_map[fn] = tgt
+
+        bootfile.write("SCRIPT_MAP=%r\n"%(script_map,))
         bootfile.write('try:\n')
-        bootfile.write('    _run(%r)\n' % os.path.realpath(script))
+        bootfile.write('    _run()\n')
         bootfile.write('except KeyboardInterrupt:\n')
         bootfile.write('    pass\n')
         bootfile.close()
@@ -1303,13 +1436,39 @@ class py2app(Command):
         target.appdir = appdir
         return appdir
 
-    def build_executable(self, target, arcname, pkgexts, copyexts, script):
+
+    def build_executable(self, target, arcname, pkgexts, copyexts, script, extra_scripts):
         # Build an executable for the target
         appdir, resdir, plist = self.create_bundle(target, script)
         self.appdir = appdir
         self.resdir = resdir
         self.plist = plist
 
+        for fn in extra_scripts:
+            if fn.endswith('.py'):
+                fn = fn[:-3]
+            elif fn.endswith('.pyw'):
+                fn = fn[:-4]
+
+            src_fn = script_executable(arch=self.arch)
+            tgt_fn = os.path.join(self.appdir, 'Contents', 'MacOS', fn)
+            mergecopy(src_fn, tgt_fn)
+            make_exec(tgt_fn)
+
+
+        site_path = os.path.join(resdir, 'site.py')
+        byte_compile([
+            SourceModule('site', site_path),
+            ],
+            target_dir=resdir,
+            optimize=self.optimize,
+            force=self.force,
+            verbose=self.verbose,
+            dry_run=self.dry_run)
+        if not self.dry_run:
+            os.unlink(site_path)
+
+
         includedir = None
         configdir = None
         if sysconfig is not None:
@@ -1335,18 +1494,175 @@ class py2app(Command):
         for fn in target.prescripts:
             bootfile.write(self.get_bootstrap_data(fn))
             bootfile.write('\n\n')
-        bootfile.write('_run(%r)\n' % (os.path.basename(script),))
+
+        bootfile.write("DEFAULT_SCRIPT=%r\n"%(os.path.basename(script),))
+        bootfile.write('try:\n')
+        bootfile.write('    _run()\n' % os.path.realpath(script))
+        bootfile.write('except KeyboardInterrupt:\n')
+        bootfile.write('    pass\n')
+        bootfile.close()
+
+        target.appdir = appdir
+        return appdir
+
+    def build_executable(self, target, arcname, pkgexts, copyexts, script, extra_scripts):
+        # Build an executable for the target
+        appdir, resdir, plist = self.create_bundle(target, script)
+        self.appdir = appdir
+        self.resdir = resdir
+        self.plist = plist
+
+        for fn in extra_scripts:
+            if fn.endswith('.py'):
+                fn = fn[:-3]
+            elif fn.endswith('.pyw'):
+                fn = fn[:-4]
+
+            src_fn = script_executable(arch=self.arch)
+            tgt_fn = os.path.join(self.appdir, 'Contents', 'MacOS', fn)
+            mergecopy(src_fn, tgt_fn)
+            make_exec(tgt_fn)
+
+
+        site_path = os.path.join(resdir, 'site.py')
+        byte_compile([
+            SourceModule('site', site_path),
+            ],
+            target_dir=resdir,
+            optimize=self.optimize,
+            force=self.force,
+            verbose=self.verbose,
+            dry_run=self.dry_run)
+        if not self.dry_run:
+            os.unlink(site_path)
+
+
+        includedir = None
+        configdir = None
+        if sysconfig is not None:
+            includedir = sysconfig.get_config_var('CONFINCLUDEPY')
+            configdir = sysconfig.get_config_var('LIBPL')
+
+
+        if includedir is None:
+            includedir = 'python%d.%d'%(sys.version_info[:2])
+        else:
+            includedir = os.path.basename(includedir)
+
+        if configdir is None:
+            configdir = 'config'
+        else:
+            configdir = os.path.basename(configdir)
+
+        self.compile_datamodels(resdir)
+        self.compile_mappingmodels(resdir)
+
+        bootfn = '__boot__'
+        bootfile = open(os.path.join(resdir, bootfn + '.py'), 'w')
+        for fn in target.prescripts:
+            bootfile.write(self.get_bootstrap_data(fn))
+            bootfile.write('\n\n')
+
+        bootfile.write("DEFAULT_SCRIPT=%r\n"%(os.path.basename(script),))
+        bootfile.write('try:\n')
+        bootfile.write('    _run()\n' % os.path.realpath(script))
+        bootfile.write('except KeyboardInterrupt:\n')
+        bootfile.write('    pass\n')
+        bootfile.close()
+
+        target.appdir = appdir
+        return appdir
+
+    def build_executable(self, target, arcname, pkgexts, copyexts, script, extra_scripts):
+        # Build an executable for the target
+        appdir, resdir, plist = self.create_bundle(target, script)
+        self.appdir = appdir
+        self.resdir = resdir
+        self.plist = plist
+
+        for fn in extra_scripts:
+            if fn.endswith('.py'):
+                fn = fn[:-3]
+            elif fn.endswith('.pyw'):
+                fn = fn[:-4]
+
+            src_fn = script_executable(arch=self.arch)
+            tgt_fn = os.path.join(self.appdir, 'Contents', 'MacOS', fn)
+            mergecopy(src_fn, tgt_fn)
+            make_exec(tgt_fn)
+
+
+        site_path = os.path.join(resdir, 'site.py')
+        byte_compile([
+            SourceModule('site', site_path),
+            ],
+            target_dir=resdir,
+            optimize=self.optimize,
+            force=self.force,
+            verbose=self.verbose,
+            dry_run=self.dry_run)
+        if not self.dry_run:
+            os.unlink(site_path)
+
+
+        includedir = None
+        configdir = None
+        if sysconfig is not None:
+            includedir = sysconfig.get_config_var('CONFINCLUDEPY')
+            configdir = sysconfig.get_config_var('LIBPL')
+
+
+        if includedir is None:
+            includedir = 'python%d.%d'%(sys.version_info[:2])
+        else:
+            includedir = os.path.basename(includedir)
+
+        if configdir is None:
+            configdir = 'config'
+        else:
+            configdir = os.path.basename(configdir)
+
+        self.compile_datamodels(resdir)
+        self.compile_mappingmodels(resdir)
+
+        bootfn = '__boot__'
+        bootfile = open(os.path.join(resdir, bootfn + '.py'), 'w')
+        for fn in target.prescripts:
+            bootfile.write(self.get_bootstrap_data(fn))
+            bootfile.write('\n\n')
+
+        bootfile.write("DEFAULT_SCRIPT=%r\n"%(os.path.basename(script),))
+
+        script_map = {}
+        for fn in extra_scripts:
+            fn = os.path.basename(fn)
+            if fn.endswith('.py'):
+                script_map[fn[:-3]] = fn
+            elif fn.endswith('.py'):
+                script_map[fn[:-4]] = fn
+            else:
+                script_map[fn] = fn
+
+        bootfile.write("SCRIPT_MAP=%r\n"%(script_map,))
+        bootfile.write('_run()\n')
         bootfile.close()
 
         self.copy_file(script, resdir)
-        pydir = os.path.join(resdir, 'lib', 'python' + sys.version[:3])
-        if sys.version_info[0] == 2:
+        for fn in extra_scripts:
+            self.copy_file(fn, resdir)
+
+        pydir = os.path.join(resdir, 'lib', 'python%s.%s'%(sys.version_info[:2]))
+        if sys.version_info[0] == 2 or self.semi_standalone:
             arcdir = os.path.join(resdir, 'lib', 'python' + sys.version[:3])
         else:
             arcdir = os.path.join(resdir, 'lib')
         realhome = os.path.join(sys.prefix, 'lib', 'python' + sys.version[:3])
         self.mkpath(pydir)
-        self.symlink('../../site.py', os.path.join(pydir, 'site.py'))
+
+        if self.optimize:
+            self.symlink('../../site.pyo', os.path.join(pydir, 'site.pyo'))
+        else:
+            self.symlink('../../site.pyc', os.path.join(pydir, 'site.pyc'))
         cfgdir = os.path.join(pydir, configdir)
         realcfg = os.path.join(realhome, configdir)
         real_include = os.path.join(sys.prefix, 'include')
@@ -1434,19 +1750,20 @@ class py2app(Command):
         pathname = os.path.join(self.temp_dir, "%s.py" % slashname)
         if os.path.exists(pathname):
             if self.verbose:
-                print ("skipping python loader for extension %r"
+                print("skipping python loader for extension %r"
                     % (item.identifier,))
         else:
             self.mkpath(os.path.dirname(pathname))
             # and what about dry_run?
             if self.verbose:
-                print ("creating python loader for extension %r"
+                print("creating python loader for extension %r"
                     % (item.identifier,))
 
             fname = slashname + os.path.splitext(item.filename)[1]
             source = make_loader(fname)
             if not self.dry_run:
-                open(pathname, "w").write(source)
+                with open(pathname, "w") as fp:
+                    fp.write(source)
             else:
                 return
         return SourceModule(item.identifier, pathname)
