@@ -28,18 +28,30 @@ import time
 import os
 import signal
 import py2app
+import hashlib
 
 DIR_NAME=os.path.dirname(os.path.abspath(__file__))
 
-if sys.version_info[0] == 2:
-    def B(value):
-        return value
+def make_checksums(path):
+    result = {}
+    for root, dnames, fnames in os.walk(path):
+        for dn in dnames:
+            result[os.path.join(root, dn)] = None
 
-else:
-    def B(value):
-        return value.encode('latin1')
+        for fn in fnames:
+            h = hashlib.sha1()
+            p = os.path.join(root, fn)
+            if os.path.islink(p):
+                result[p] = os.readlink(p)
 
+            else:
+                with open(p, 'rb') as fp:
+                    block = fp.read(10240)
+                    while block:
+                        h.update(block)
+                        block = fp.read(10240)
 
+                result[p] = h.hexdigest()
 
 
 class TestBasicApp (unittest.TestCase):
@@ -59,7 +71,7 @@ class TestBasicApp (unittest.TestCase):
             env['PYTHONPATH'] = pp + ':' + env['PYTHONPATH']
         else:
             env['PYTHONPATH'] = pp
-	
+
         if 'LANG' not in env:
             # Ensure that testing though SSH works
             env['LANG'] = 'en_US.UTF-8'
@@ -78,6 +90,14 @@ class TestBasicApp (unittest.TestCase):
             print (lines)
             raise AssertionError("Creating basic_app bundle failed")
 
+        cls.checksums = make_checksums(
+                os.path.join(cls.app_dir, 'dist/BasicApp.app'))
+
+    def assertChecksumsSame(self):
+        self.assertEqual(self.checksums,
+            make_checksums(
+                os.path.join(self.app_dir, 'dist/BasicApp.app')))
+
     @classmethod
     def tearDownClass(cls):
         if os.path.exists(os.path.join(cls.app_dir, 'build')):
@@ -126,6 +146,8 @@ class TestBasicApp (unittest.TestCase):
 
         p.stdout.close()
 
+        self.assertChecksumsSame()
+
     def test_simple_imports(self):
         p = self.start_app()
 
@@ -133,13 +155,13 @@ class TestBasicApp (unittest.TestCase):
         p.stdin.write('import_module("os")\n'.encode('latin1'))
         p.stdin.flush()
         ln = p.stdout.readline()
-        self.assertEqual(ln.strip(), B("os"))
+        self.assertEqual(ln.strip(), b"os")
 
         # Dependency of the main module:
         p.stdin.write('import_module("decimal")\n'.encode('latin1'))
         p.stdin.flush()
         ln = p.stdout.readline()
-        self.assertEqual(ln.strip(), B("decimal"))
+        self.assertEqual(ln.strip(), b"decimal")
 
         can_import_stdlib = False
         if '--alias' in self.py2app_args:
@@ -162,14 +184,14 @@ class TestBasicApp (unittest.TestCase):
             p.stdin.write('import_module("xdrlib")\n'.encode('latin1'))
             p.stdin.flush()
             ln = p.stdout.readline()
-            self.assertEqual(ln.strip(), B("xdrlib"))
+            self.assertEqual(ln.strip(), b"xdrlib")
 
         if sys.prefix.startswith('/System'):
             # py2app is included as part of the system install
             p.stdin.write('import_module("py2app")\n'.encode('latin1'))
             p.stdin.flush()
             ln = p.stdout.readline().decode('utf-8')
-            self.assertEqual(ln.strip(), B("py2app"))
+            self.assertEqual(ln.strip(), b"py2app")
 
 
         else:
@@ -181,6 +203,21 @@ class TestBasicApp (unittest.TestCase):
 
         p.stdin.close()
         p.stdout.close()
+        self.assertChecksumsSame()
+
+    def test_is_optimized(self):
+        p = self.start_app()
+
+        try:
+            p.stdin.write('print(__debug__)\n'.encode('latin1'))
+            p.stdin.flush()
+            ln = p.stdout.readline()
+            self.assertEqual(ln.strip(), b"True")
+
+        finally:
+            p.stdin.close()
+            p.stdout.close()
+        self.assertChecksumsSame()
 
 class TestBasicAliasApp (TestBasicApp):
     py2app_args = [ '--alias', ]
@@ -227,24 +264,38 @@ class TestBasicSemiStandaloneAppUnicodePath (TestBasicAppUnicodePath):
 class TestOptimized1 (TestBasicApp):
     py2app_args = [ '-O1' ]
 
-    def testIsOptimized(self):
+    def test_is_optimized(self):
         p = self.start_app()
 
-        p.stdin.write('print(__debug__)\n'.encode('latin1'))
-        p.stdin.flush()
-        ln = p.stdout.readline()
-        self.assertEqual(ln.strip(), B("False"))
+        try:
+            p.stdin.write('print(__debug__)\n'.encode('latin1'))
+            p.stdin.flush()
+            ln = p.stdout.readline()
+            self.assertEqual(ln.strip(), b"False")
+
+        finally:
+            p.stdin.close()
+            p.stdout.close()
+
+        self.assertChecksumsSame()
 
 class TestOptimized2 (TestBasicApp):
     py2app_args = [ '-O2' ]
 
-    def testIsOptimized(self):
+    def test_is_optimized(self):
         p = self.start_app()
 
-        p.stdin.write('print(__debug__)\n'.encode('latin1'))
-        p.stdin.flush()
-        ln = p.stdout.readline()
-        self.assertEqual(ln.strip(), B("False"))
+        try:
+            p.stdin.write('print(__debug__)\n'.encode('latin1'))
+            p.stdin.flush()
+            ln = p.stdout.readline()
+            self.assertEqual(ln.strip(), b"False")
+
+        finally:
+            p.stdin.close()
+            p.stdout.close()
+
+        self.assertChecksumsSame()
 
 if __name__ == "__main__":
     unittest.main()
